from Pypeline.Pypes.StringSearch.BoyerMoore import *
import unittest

class BoyerMooreTest(unittest.TestCase):
	"""
	Tests for functions in the BoyerMoore module.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO

	def test_search(self):
		txt = "fjkasjejaw; lcfjaxkiwaopedask;lcas;ld"
		pat = ";lcas"
		pat1 = ";"
		pat2 = "@"
		pat3 = " "
		self.assertEqual(search(txt, pat),29)
		self.assertEqual(search(txt, pat1),10)
		self.assertEqual(search(txt, pat2),None)
		self.assertEqual(search(txt, pat3),11)



from Pypeline.Pypes import MinHeap

import unittest

import numpy as np

class MinHeapTest(unittest.TestCase):
	"""
	Tests for methods in the MinHeap class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO

	def test_heapify(self):
		li = [2, 4 , 10, 8, 90, 65, 18]
		heap = MinHeap.MinHeap()
		heap.heapify(li)
		assert len(heap.MinHeap) == 7
		assert heap.MinHeap == [90, 8, 65, 2, 4, 10, 18]

	def test_heappop(self):
		li = [2, 4 , 10, 8, 90, 65, 18]
		heap = MinHeap.MinHeap()
		heap.heapify(li)
		assert len(heap.MinHeap) == 7
		assert heap.heappop() == 90

	def test_heappush(self):
		li = [2, 4, 10, 90, 65, 18]
		heap = MinHeap.MinHeap()
		heap.heapify(li)
		assert len(heap.MinHeap) == 6
		heap.heappush(8)
		assert len(heap.MinHeap) == 7
		assert heap.MinHeap == [90, 65, 18, 4, 2, 10, 8]

	def test_to_array(self):
		li = [2, 4 , 10, 8, 90, 65, 18]
		heap = MinHeap.MinHeap()
		heap.heapify(li)
		assert (heap.to_array() == np.array([90, 8, 65, 2, 4, 10, 18])).all()

	def test_to_list(self):
		li = [2, 4 , 10, 8, 90, 65, 18]
		heap = MinHeap.MinHeap()
		heap.heapify(li)
		assert heap.to_list() == [90, 8, 65, 2, 4, 10, 18]

	def test_merge(self):
		li = [2, 4 , 10]
		li2 = [8, 90, 65, 18]
		heap = MinHeap.MinHeap()
		heap.heapify(li)
		assert len(heap.MinHeap) == 3
		heap.merge(li2)
		assert len(heap.MinHeap) == 7
		assert heap.MinHeap == [90, 10, 65, 8, 4, 2, 18]


class BackwardsTest(unittest.TestCase):
	"""
	Tests for methods in the Backwards class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO


class RestoreTest(unittest.TestCase):
	"""
	Tests for methods in the Restore class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO




"""
Implementation of the heap

"""

from typing import ClassVar, List

class Heap(object):
    """ 
    Implementation of a Heap (https://en.wikipedia.org/wiki/Heap_(data_structure))
      
    Attributes:\n
        heap: the underlying list data sturcture for the heap
        currentPosition: used to keep track of the index in the heap
    """

    # Class constant variables can be assigned this way
    HEAP_SIZE: ClassVar[int] = 10

    def __init__(self, size: int=10) -> None:
        """ 
        Constructor of the Heap class

        Parameters:\n
            size: desired size of the Heap
        """
        Heap.HEAP_SIZE = size
        self.heap: List[int] = [0]*Heap.HEAP_SIZE
        self.currentPosition: int = -1

    def insert(self, item: int) -> None:
        """ 
        Public method used to insert an item into the Heap

        Parameters:\n
            item: the data to be added to the heap
        """
        if self.isFull():
            print("Heap is full...")
            return

        else:
            self.currentPosition += 1
            self.heap[self.currentPosition] = item
            self.__fixUp(self.currentPosition) # reheap the heap

    def heapSort(self) -> None:
        """ 
        Public method used to sort the heap
        """
        for i in range(self.currentPosition+1):
            temp = self.heap[0]
            print("%d" % temp)
            self.heap[0] = self.heap[self.currentPosition - i]
            self.heap[self.currentPosition] = temp
            # fix down 
            self.__fixDown(0,self.currentPosition-i-1)

    def isFull(self) -> bool:
        """ 
        Public method used to determine if the heap is full
        """
        if self.currentPosition + 1 == Heap.HEAP_SIZE:
            return True
        else:
            return False

    def __fixUp(self, index: int) -> None:
        """ 
        Private method used to reheap the heap during insertion

        Parameters:\n
            index: the index of the newly added data
        """
        parentIndex = int((index-1)/2)

        while parentIndex >= 0 and self.heap[parentIndex] < self.heap[index]:
            temp = self.heap[parentIndex]
            self.heap[parentIndex] = self.heap[index]
            self.heap[index] = temp
            parentIndex = int((index-1)/2) # from current move up

    def __fixDown(self, index: int, upto: int) -> None:
        """ 
        Private method used to reheap the heap during heap sort

        Parameters:\n
            index: the index of the newly added data
            upto: end index for the heap down operation
        """
        while index <= upto:
            leftChild = 2*index +1
            rightChild = 2*index +2

            if leftChild < upto:
                childToSwap = None

                if rightChild > upto:
                    childToSwap = leftChild
                else:
                    if self.heap[leftChild] > self.heap[rightChild]:
                        childToSwap = leftChild
                    else:
                        childToSwap = rightChild
                if self.heap[index] < self.heap[childToSwap]:
                    temp = self.heap[index]
                    self.heap[index] = self.heap[childToSwap]
                    self.heap[childToSwap] = temp
                else:
                    break

                index = childToSwap
            else:
                break



from Pypeline.Pypes.DoublyLinkedList import DoublyLinkedList, Node
import unittest

class NodeTest(unittest.TestCase):
	"""
	Tests for methods in the Node class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO


class DoublyLinkedListTest(unittest.TestCase):
	"""
	Tests for methods in the DoublyLinkedList class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		self.dll = DoublyLinkedList()

	def tearDown(self):
		self.dll = None

	def test_push(self):
		self.dll.push(0)
		assert self.dll.head.data == 0
		[self.dll.push(x) for x in range(1, 10)]
		node = self.dll.head
		i = 9
		while node:
			print(node.data)
			assert(node.data == i)
			i -= 1
			node = node.next

	def test_insertAfter(self):
		self.dll.push(1)
		self.dll.push(3)
		self.dll.insertAfter(self.dll.head, 2)
		node = self.dll.head
		i = 3
		while node:
			print(node.data)
			assert(node.data == i)
			i-=1
			node = node.next


	def test_append(self):
		self.dll.push(3)
		self.dll.push(1)
		self.dll.insertAfter(self.dll.head, 2)
		self.dll.append(4)
		node = self.dll.head
		i = 1
		while node:
			print(node.data)
			assert(node.data == i)
			i+=1
			node = node.next



from Pypeline.Pypes.Node import Node
import unittest

class NodeTest(unittest.TestCase):
	"""
	Tests for methods in the Node class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		self.node = Node()

	def tearDown(self):
		self.node = None

	def test_setVisited(self):
		self.node.setVal(10)
		self.node.setVisited()
		assert self.node.isVisited() == True

	def test_swapVisited(self):
		self.node.swapVisited()
		assert self.node.isVisited()
		self.node.setVal(10)
		self.node.setVisited()
		self.node.swapVisited()
		assert self.node.isVisited() == False

	def test_isVisited(self):
		assert self.node.isVisited() == False

	def test_setVal(self):
		self.node.setVal(10)
		assert self.node.getVal() == 10

	def test_getVal(self):
		assert self.node.getVal() == None
		self.node.setVal(10)
		assert self.node.getVal() == 10



"""
Implementation of the AVL tree !!

"""

from typing import Any, Optional

class Node(object):
    """ 
    This is a custom node class for the AVL Tree

    Attributes:\n
        data: The data to be stored in the node
        rightChild: The right child of the current node
        leftChild: The left child of the current node
        height: The length of longest path from current not to lead. Used to check if the tree is balanced. 
    """

    def __init__(self, data: Any) -> None:
        """ 
        Constructor of the Node class

        Parameters:\n
            data: the data to be stored in the Node 
        """
        self.data: Any = data
        self.rightChild: Optional[Node] = None
        self.leftChild: Optional[Node] = None
        self.height: int = 0
        

class AVL(object):
    """ 
    Implementation of the AVL Tree (https://en.wikipedia.org/wiki/AVL_tree)
    
    Attributes:\n
        root: The root node of the AVL Tree
    """

    def __init__(self) -> None:
        """ 
        Constructor of the AVL class
        """
        self.root: Optional[Node] = None

    def calcHeight(self, node: Optional[Node]) -> int:
        """ 
        Method used to determine the height of a particular node

        Parameters:\n
            node: the node for which the height is to be calculated
        """
        if not node: # if the node is None
            return -1
        else:
            return node.height

    # if return value > 1, it means left heavy situations -> right rotation
    # if return value < -1, it means right heavy situations -> left rotation
    # if between -1 and 1, it is balanced

    def calcBalance(self, node: Optional[Node]) -> int:
        """ 
        Method used to determine whether a particular node is balanced

        if return value > 1, it means left heavy situations -> right rotation\n
        if return value < -1, it means right heavy situations -> left rotation\n
        otherwise it is balanced\n

        Parameters:\n
            node: the node that will have its level of balance calculated
        """
        if not node:
            return 0
        else: 
            return self.calcHeight(node.leftChild) - self.calcHeight(node.rightChild)

    def rotateRight(self, node: Node) -> Node:
        """ 
        Method used to perform a right rotation from a particular node.


        Rotations to the right and the left are symmetrical operations. Rotation
        operations are quite fast as it it just updating references O(1) time complexity.

        Parameters:\n
            node: the node that will be reference rotatation point
        """
        tempLeft = node.leftChild
        t = tempLeft.rightChild

        tempLeft.rightChild = node
        node.leftChild = t

        # update node's height by checking left and right sub tree
        node.height = max(self.calcHeight(node.leftChild),self.calcHeight(node.rightChild)) + 1
        tempLeft.height = max(self.calcHeight(tempLeft.leftChild),self.calcHeight(tempLeft.rightChild)) + 1

        # return the new root node
        return tempLeft

    def rotateLeft(self, node: Node) -> Node:
        """ 
        Method used to perform a left rotation from a particular node.

        Rotations to the right and the left are symmetrical operations. Rotation
        operations are quite fast as it it just updating references O(1) time complexity.

        Parameters:\n
            node: the node that will be reference rotatation point
        """
        tempRight = node.rightChild
        t = tempRight.leftChild

        tempRight.leftChild = node
        node.rightChild = t

        # update node's height by checking left and right sub tree
        node.height = max(self.calcHeight(node.leftChild),self.calcHeight(node.rightChild)) + 1
        tempRight.height = max(self.calcHeight(tempRight.leftChild),self.calcHeight(tempRight.rightChild)) + 1

        # return the new root node
        return tempRight

    def insert(self, data: Any) -> None:
        """ 
        Public method used to insert data into the AVL tree

        A check is necessary to ensure the AVL property is not violated. This
        check is performed within the associated helper method `__insertNode`

        Parameters:\n
            data: the data to be added to the tree
        """
        self.root = self.__insertNode(data, self.root)

    def __insertNode(self, data: Any, node: Optional[Node]) -> Node:
        """ 
        Private method used to insert data into the AVL tree as well as
        check if the AVL property has been violated.

        Parameters:\n
            data: the data to be added to the tree
            node: the root node of the tree
        """
        if not node: # base case so plug in the value
            return Node(data)
        else:
            if data < node.data: # call recursively on left subtree to insert
                node.leftChild = self.__insertNode(data, node.leftChild)
            else: # call recursively on right subtree
                node.rightChild = self.__insertNode(data, node.rightChild)

            # update height parameter
            node.height =  max(self.calcHeight(node.leftChild), self.calcHeight(node.rightChild)) + 1

            # fixes any errors in the balance of the tree
            return self.__settleViolation(data,node)

    def __settleViolation(self, data: Any, node: Node) -> Node:
        """ 
        Method used to settle AVL violations as they arrise.

        Parameters:\n
            data: the data to be added to the tree
            node: the root node of the tree
        """
        balance = self.calcBalance(node)
        # Case 1: Greater than 1 -> left left heavy -> & current data is smaller than left child so -> simple right rotation
        if balance > 1 and data < node.leftChild.data:
            #print("Doubly left heavy situation...")
            return self.rotateRight(node)

        # Case 2: Less than -1 -> right right heavy -> & current data is smaller than right child  so -> simple left rotation
        elif balance < -1 and data > node.rightChild.data:
            #print("Doubly right heavy situation...")
            return self.rotateLeft(node)
        # Case 3: Left right
        elif balance > 1 and data > node.leftChild.data: # current data is greater than left (so its on right)
            #print("Left-right heavy situation...")
            node.leftChild = self.rotateLeft(node.leftChild)
            return self.rotateRight(node)
        # Case 4: Right Left
        elif balance < -1 and data < node.rightChild.data:
            #print("Right-left heavy situation...")
            node.rightChild = self.rotateRight(node.rightChild)
            return self.rotateLeft(node)

        return node

    def traverse(self) -> None:
        """ 
        Method used to perform an inOrderTraversal
        """
        if self.root:
            self.__inOrderTraversal(self.root)
    
    def __inOrderTraversal(self, node: Node) -> None:
        """ 
        Method used to perform an inOrderTraversal

        Parameters:\n
            node: the root node of the tree
        """
        if node.leftChild:
            self.__inOrderTraversal(node.leftChild)

        print("%s" % node.data)

        if node.rightChild:
            self.__inOrderTraversal(node.rightChild)

    def remove(self, data: Any) -> None:
        """ 
        Public method used to remove some piece of data from the tree

        Parameters:\n
            data: the data to be removed to the tree
        """
        if self.root: # if the root exists -> there are nodes
            self.root = self.__removeData(data, self.root)

    def __removeData(self, data: Any, node: Optional[Node]) -> Optional[Node]:
        """ 
        Private method used to remove some piece of data from the tree.

        The predecessor method of removing data is used.

        Parameters:\n
            data: the data to be removed to the tree
            node: the root node of the tree
        """
        if not node: # this is the base case my guy duuhh
            return node

        if data < node.data: # if smaller
            node.leftChild =  self.__removeData(data, node.leftChild)
        elif data > node.data:
            node.rightChild = self.__removeData(data, node.rightChild)
        else:
            # 1: if there are no children -> Leaf Node
            if not node.leftChild and not node.rightChild:
                #print("Removing leaf node...")
                del node
                return None
            elif not node.leftChild: # no left but would have right ^
                #print("Removing node with right child...")
                rightChild = node.rightChild
                del node
                return rightChild
            elif not node.rightChild: # no right child
                #print("Removing node with left child...")
                leftChild = node.leftChild
                del node
                return leftChild
            else:
                #print("Removing node with two children...")
                predecessor = self.__getPredecessor(node.leftChild)
                node.data = predecessor.data
                node.leftChild = self.__removeData(predecessor.data, node.leftChild)

        if not node:
            return node

        node.height = max(self.calcHeight(node.leftChild),self.calcHeight(node.rightChild)) + 1
        
        balance = self.calcBalance(node)

        # Case 1: Greater than 1 -> left left heavy -> & current data is smaller than left child so -> simple right rotation
        if balance > 1 and self.calcBalance(node.leftChild) >= 0:
            #print("Doubly left heavy situation...")
            return self.rotateRight(node)

        # Case 2: Less than -1 -> right right heavy -> & current data is smaller than right child  so -> simple left rotation
        elif balance < -1 and self.calcBalance(node.rightChild) <= 0:
            #print("Doubly right heavy situation...")
            return self.rotateLeft(node)
        # Case 3: Left right
        elif balance > 1 and self.calcBalance(node.leftChild) < 0: # current data is greater than left (so its on right)
            #print("Left-right heavy situation...")
            node.leftChild = self.rotateLeft(node.leftChild)
            return self.rotateRight(node)
        # Case 4: Right Left
        elif balance < -1 and self.calcBalance(node.rightChild) > 0:
            #print("Right-left heavy situation...")
            node.rightChild = self.rotateRight(node.rightChild)
            return self.rotateLeft(node)

        return node                 

    def __getPredecessor(self, node: Node) -> Node:
        """ 
        Private method to get the largest value in a subtree

        Parameters:\n
            node: the reference node used to find largest value in its subtree
        """
        if not node.rightChild:
            return node
        return self.__getPredecessor(node.rightChild)


"""
Implementation of the linked list

"""

from typing import Any, List, Optional

import numpy as np

class Node(object):
    """ 
    This is a custom node class for the Linked List
      
    Attributes:\n
        data: The data to be stored in the node
        nextNode: Reference to the next node
    """

    def __init__(self, data: Any) -> None: # instantiates data in the node
        self.data: None = data
        self.nextNode: Optional[Node] = None # reference to the next node

class LinkedList(object):
    """
    Implementation of the linked list (https://en.wikipedia.org/wiki/Linked_list)

    """

    def __init__(self) -> None:
        self.head: Optional[Node] = None # identiify head of the list
        self.size: int = 0 # initialize length to be zero
        self.linkedlist: List[Any] = []

    def __str__(self) -> str:
        return str(self.linkedlist)

    def __repr__(self) -> str:
        return str(self.linkedlist)

    def insert(self, data: Any) -> None:
        self.size += 1
        #instantiate a new node
        newNode = Node(data)

        # if head of ll not defined, set it
        if not self.head:
            self.head = newNode
        # if it is, this new node becomes head of ll
        else: # essentially just updating pointers
            newNode.nextNode = self.head
            self.head = newNode
        self.linkedlist.append(data)

    # insert item to the end of the list; O(N) time complexity
    def insertEnd(self, data: Any) -> None:
        self.size +=1
        newNode = Node(data)
        #print(newNode,newNode.nextNode)
        actualNode = self.head
        #print(actualNode,actualNode.nextNode)

        # while the following node is not None (not the end of ll)
        while actualNode.nextNode is not None:
            #print(actualNode.nextNode)
            actualNode = actualNode.nextNode # increment
            # brings us to last item

        actualNode.nextNode = newNode
        self.linkedlist.append(data)

    # checks size of the ll; O(1) time complexity
    def getSize(self) -> int: 
        return self.size

    # calculates the size of the ll O(N) time complexity; not very efficient
    def getSize2(self) -> int:
        actualNode = self.head
        size = 0

        while actualNode is not None:
            # increment the size, and shift to next node in the ll
            size +=1
            actualNode = actualNode.nextNode
        return size

    def traverseList(self) -> None:
        actualNode = self.head

        while actualNode is not None:
            # print current data
            print("%d " % actualNode.data)
            # go to he next node in ll
            actualNode = actualNode.nextNode

    # define remove method for particular data O(N) time complexity
    def remove(self, data: Any) -> None:
        # if there is no head in the list, then cant create a list
        if self.head is None:
            return
        self.size -= 1

        currentNode = self.head
        previousNode = None

        while currentNode.data != data:
            previousNode = currentNode
            currentNode = currentNode.nextNode

        # at this point, we are at the node we want to get rid of
        if previousNode is None: # if we are at the head
            self.head = currentNode.nextNode

        else: # if not at the head
            # change the pointer
            previousNode.nextNode = currentNode.nextNode

        self.linkedlist.pop(self.linkedlist.index(data))

    def to_array(self) -> Any:
        return np.array(self.linkedlist)

    def to_list(self) -> List[Any]:
        return self.linkedlist



from Pypeline.Pypes.Sorting import Radixsort

import unittest

class RadixsortTest(unittest.TestCase):
	"""
	Tests for functions in the Radixsort module.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO

	def test_countingSort(self):
		arr = [3, 2, 2, 1, 4]
		Radixsort.countingSort(arr, 1)
		assert arr == [1, 2, 2, 3, 4]

	def test_radixSort(self):
		arr = [3, 2, 2, 1, 4]
		Radixsort.radixSort(arr)
		assert arr == [1, 2, 2, 3, 4]



from Pypeline.Pypes.Sorting import Helper
import unittest

class HelperTest(unittest.TestCase):
	"""
	Tests for functions in the Helper module.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO

	def test_less(self):
		assert Helper.less(5, 3) == False
		assert Helper.less(3, 3) == False
		assert Helper.less(1, 3) == True

	def test_swap(self):
		arr = [1, 2]
		Helper.swap(arr, 0, 1)
		assert arr == [2, 1]
		Helper.swap(arr, 1, 1)
		assert arr == [2, 1]


"""
Implementation of the node object

"""

from typing import Any, Optional

class Node(object):

    def __init__(self, val: Any=None):
        '''
        Base for node object containing node value and visited state.
        val -> Any data type
        visited -> bool
        '''
        self.val: Any = val
        self.visited: bool = False

    def setVisited(self) -> None:
        self.visited = True

    def swapVisited(self) -> None:
        self.visited = not self.visited

    def isVisited(self) -> bool:
        return self.visited

    def setVal(self, val) -> None:
        self.val = val

    def getVal(self) -> Any:
        return self.val


from Pypeline.Pypes.Sorting import Mergesort
import unittest

class MergesortTest(unittest.TestCase):
	"""
	Tests for functions in the Mergesort module.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO

	def test_merge(self):
		arr = [1, 4, 2, 8, 7]
		Mergesort.merge(arr, 0, 1, 2, order='ascend')
		assert arr == [1, 2, 4, 8, 7]

	def test_mergeSort(self):
		arr = [1, 4, 2, 8, 7]
		Mergesort.mergeSort(arr, 0, 4)
		assert arr == [1, 2, 4, 7, 8]



"""
Implementation of the Stack Data Structure

N.B:
    - Array representation -> LIFO Structure

"""

from typing import Any, List, Optional

import numpy as np

class Stack:
    """
    Implementation of the Stack Data Structure (https://en.wikipedia.org/wiki/Stack_(abstract_data_type))

    An abstract data type that serves as a collection of elements, with two main principal operations: push and pop
    from the top of the list (most recent values).
    """

    def __init__(self) -> None:
        self.stack: List[Any] = []

    def isEmpty(self) -> bool:
        return self.stack == []

    def sizeStack(self) -> int:
        return len(self.stack)

    # 3 main methods in a stack! Push, Pop and Peek
    def push(self, data: Any) -> None:
        self.stack.append(data)

    def pop(self) -> Any:
        data = self.stack[-1]
        #self.stack = self.stack[:-1]
        del self.stack[-1]
        return data

    def peek(self) -> Any:
        return self.stack[-1]

    def to_array(self) -> Any:
        return np.array(self.stack)

    def to_list(self) -> List[Any]:
        return self.stack




from Pypeline.Pypes.Trie import Trie
import pytest
import unittest

class NodeTest(unittest.TestCase):
	"""
	Tests for methods in the Node class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO


class TrieTest(unittest.TestCase):
	"""
	Tests for methods in the Trie class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		self.tree = Trie()

	def tearDown(self):
		self.tree = None

	def test_insert(self):
		self.tree.insert('test')
		assert self.tree.search("test")

	def test_search(self):

		self.tree.insert('hack')
		assert self.tree.search("hack")

		self.tree.insert('hackathon')
		assert self.tree.search("hackathon")

		self.tree.insert('hac')
		assert self.tree.search("hac")

		assert self.tree.search("ha") == False
		assert self.tree.search("h") == False

		with pytest.raises(TypeError):
			self.tree.insert()
			self.tree.insert(12)
			self.tree.insert(['test'])


from Pypeline.Pypes.Heap import Heap
import unittest

class HeapTest(unittest.TestCase):
	"""
	Tests for methods in the Heap class.
	"""

	@classmethod
	def setUpClass(cls):
		cls.vals = [10, 0, 2, 40, 70, -20]

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		self.heap = Heap(5)

	def tearDown(self):
		self.heap = None

	def test_insert(self):
		[self.heap.insert(x) for x in self.vals[:3]]
		assert not self.heap.isFull()

	def test_heapSort(self):
		[self.heap.insert(x) for x in self.vals[:-1]]
		self.heap.heapSort()
		assert self.heap.heap

	def test_isFull(self):
		[self.heap.insert(x) for x in self.vals]
		assert self.heap.isFull()



"""
Implementation of the Trie Data Structure

"""

from typing import Dict

class Node(object):
    """ 
    This is a custom node class for the Trie
      
    Attributes:\n
        char: The char to be stored
        children: Dictionary of the nodes children
        word_finished: Inidicate whether the word is finished at this node
        counter: number of word occurences using this node
    """

    def __init__(self,char) -> None:
        self.char: str = char
        self.children: Dict[str, Node] = {}
        self.word_finished: bool = False
        self.counter: int = 0

class Trie(object):
    """
    Implementation of the Trie Data Structure (https://en.wikipedia.org/wiki/Trie)

    A type of search tree, a tree data structure used for locating specific keys from within a set

    """

    def __init__(self) -> None:
        self.root: Node = Node('*')

    def insert(self, word: str) -> None:
        if type(word) != str:
            raise TypeError(f"Object of type {type(word)} cannot be inserted into trie since only str is supported")

        current = self.root
        for char in word:
            if char in current.children: # essentially counting char occurences
                current = current.children[char]
                current.counter += 1
            else: # if not already defined, generate a new node
                new_node = Node(char)
                current.children[char] = new_node
                current = new_node
                current.counter += 1
        current.word_finished = True # when finished flag!! 

    def search(self, word: str) -> bool:
        if not self.root.children: # if there are no children of the root, no words to search
            return False
        current = self.root
        for char in word:
            if char in current.children:
                current = current.children[char]
            else:
                return False
        if current.word_finished:
            return True
        return False



from Pypeline.Pypes.AVLTree import AVL, Node
import unittest

class NodeTest(unittest.TestCase):
	"""
	Tests for methods in the Node class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO


class AVLTest(unittest.TestCase):
	"""
	Tests for methods in the AVL class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		self.avl = AVL()

		# self.avl.traverse()
		# self.avl.remove(15)
		# self.avl.remove(20)
		# self.avl.traverse()
		pass #TODO

	def tearDown(self):
		self.avl = None

	def test_AVL_rebalance(self):
		# This test will check for both left and right rotations with a simple 3-node tree
		# TODO: LL and RR rotations
		self.avl.insert(20)
		assert not(self.avl.root.leftChild and self.avl.root.rightChild)
		self.avl.insert(4)
		assert self.avl.root.leftChild.data == 4
		self.avl.insert(15)
		assert self.avl.root.data == 15 and self.avl.root.rightChild.data == 20
		self.avl.remove(15)
		assert self.avl.root.data == 4 and self.avl.root.rightChild.data == 20
		self.avl.remove(4)
		self.avl.insert(4)
		assert self.avl.root.data == 20 and self.avl.root.leftChild.data == 4
		self.avl.insert(8)
		assert self.avl.root.data == 8 and self.avl.root.leftChild.data == 4



from Pypeline.Pypes.LinkedList import LinkedList, Node
import numpy as np
import unittest

class NodeTest(unittest.TestCase):
	"""
	Tests for methods in the Node class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO


class LinkedListTest(unittest.TestCase):
	"""
	Tests for methods in the LinkedList class.
	"""

	@classmethod
	def setUpClass(cls):
		cls.vals = [12, 3, 78, 5]

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		self.ll = LinkedList()

		# linkedlist.traverseList()
		# print("Size of the linked list is {}.".format(linkedlist.getSize()))

		# linkedlist.remove(5)
		# linkedlist.remove(12)
		# linkedlist.remove(3)
		# print("Items deleted from list.")
		# linkedlist.traverseList()
		# linkedlist.remove(78)
		# linkedlist.traverseList()
		# print("Size of linkedlist now is %d." % linkedlist.getSize())

	def tearDown(self):
		self.ll = None

	def test_insert(self):
		[self.ll.insert(x) for x in self.vals[:-1]]
		assert self.ll.size == len(self.vals) - 1 

	def test_insertEnd(self):
		[self.ll.insert(x) for x in self.vals[:-1]]
		self.ll.insertEnd(self.vals[-1])
		node = self.ll.head
		while node.nextNode:
			node = node.nextNode
		assert node.data == self.vals[-1]

	def test_getSize(self):
		[self.ll.insert(x) for x in self.vals]
		assert self.ll.getSize() == len(self.vals)

	def test_getSize2(self):
		[self.ll.insert(x) for x in self.vals]
		assert self.ll.getSize2() == len(self.vals)

	def test_remove(self):
		[self.ll.insert(x) for x in self.vals]
		self.ll.remove(78)
		self.ll.remove(12)
		assert self.ll.head.data == 5
		assert self.ll.size == 2


	def test_to_array(self):
		[self.ll.insert(x) for x in self.vals]
		arr = self.ll.to_array()
		assert type(arr) == np.ndarray
		assert (np.array(self.vals) == arr).all()

	def test_to_list(self):
		[self.ll.insert(x) for x in self.vals]
		arr = self.ll.to_list()
		assert type(arr) == list
		assert self.vals == arr




class Node:
    """ 
    This is a custom node class for the Red Black BST
      
    Attributes:\n
        key: the key used for comparison
        size: the size of the tree
        color: True indicates Red and False indicate a black link
        val: the data stored in the node
        left: Left node child
        right: Right node child
    """

    def __init__(self, key, val, size, color):
        self.key = key
        self.val = val
        self.N = size
        self.color = color
        self.left = None
        self.right = None


def isRed(x):
    """ 
    A helper function to determine if the link is red or not

    Parameters:\n
        x: a node in the tree
    """
    if x is None:
        return False
    return x.color == True


def size(root):
    """ 
    A helper function to determine the size of tree

    Parameters:\n
        root: a root in the tree
    """
    if root == None:
        return 0
    return root.N


def rotateLeft(h):
    """ 
    A helper function to restore the red-black tree order by changing the color of each node and added their sizes together

    Parameters:\n
        h: a node in the tree
    """
    x = h.right
    h.right = x.left
    x.left = h
    x.color = h.color
    h.color = True
    x.N = h.N
    h.N = 1 + size(h.left) + size(h.right)
    return x


def rotateRight(h):
    """ 
    A helper function to restore the red-black tree order by changing the color of each node and added their sizes together

    Parameters:\n
        h: a node in the tree
    """
    x = h.left
    h.left = x.right
    x.right = h
    x.color = h.color
    h.color = True
    x.N = h.N
    h.N = 1 + size(h.left) + size(h.right)
    return x

def flipColors(h):
    """ 
    A helper function to flip the colors of the tree by changing the color of each node

    Parameters:\n
        h: a node in the tree
    """
    h.color = True
    h.left.color = False
    h.left.color = False


def cmp(v1, v2):
    """ 
    A helper function to compare the two values

    Parameters:\n
        v1: first value
        v2: second value
    """
    if v1 < v2:
        return -1
    elif v1 == v2:
        return 0
    else:
        return 1


class redBlackBST:
    """
    Implementation of the Red Black BST

    Attributes:\n
        root: value is set to None first unless Put is called
    """

    ## @brief constructor for the redBlackBST
    #  @details root value is set to None first unless Put is called
    #  @param self the self parameter
    def __init__(self):
        self.root = None

    ## @brief A method that puts the node into the BST
    #  @details calling another helper method to finish the heavy lifting work
    #  @param key the key used for comparison
    #  @param val the value of the node being put
    def put(self, key, val):
        """ 
        A method that puts the node into the BST calling another helper method to finish the heavy lifting work

        Parameters:\n
            key: the key used for comparison
            val: the value of the node being put
        """
        self.root = self.actualPut(self.root, key, val)
        self.root.color = False

    ## @brief A method that puts the node into the BST
    #  @details calling different helper function to ensures the Red black order is preserved
    #  @param key the key used for comparison
    #  @param val the value of the node being put
    #  @param h the node we want to put at

    def actualPut(self, h, key, val) -> Node:
        """ 
        A method that puts the node into the BST calling different helper function to ensures the Red black order is preserved

        Parameters:\n
            key: the key used for comparison
            val: the value of the node being put
            h: the node we want to put at
        """
        if h is None:
            return Node(key, val, 1, True)
        compare = cmp(key, h.key)
        if compare < 0:
            h.left = self.actualPut(h.left, key, val)
        elif compare > 0:
            h.right = self.actualPut(h.right, key, val)
        else:
            h.val = val
        if isRed(h.right) and not isRed(h.left):
            h = rotateLeft(h)
        if isRed(h.left) and isRed(h.left.left):
            h = rotateRight(h)
        if isRed(h.left) and isRed(h.right):
            flipColors(h)
        h.N = size(h.left) + size(h.right) + 1
        return h

    ## @brief A getter for the BST
    #  @details Comparing the key values in logarithmic time to find the node
    #  @param key the key used for comparison

    def get(self, key):
            cur = self.root
            while cur is not None:
                if cur.key == key:
                    return cur
                elif cmp(key, cur.key) < 0:
                    cur = cur.left
                else:
                    cur = cur.right
            return None
    def inorder(self,root):
        if root is None:
            return
        self.inorder(root.left)
        print(root.val)
        self.inorder(root.right)


"""
Implementation of the binary search tree

"""

from typing import Optional, Union

class Node(object):
    """ 
    This is a custom node class for the Binary Search Tree
      
    Attributes:\n
        data: The data to be stored in the node
        rightChild: The right child of the current node
        leftChild: The left child of the current node 
    """

    def __init__(self, data: Union[int, float]) -> None:
        """ 
        Constructor of the Node class

        Parameters:\n
            data: the data to be stored in the Node 
        """
        assert (isinstance(data, int) or isinstance(data,float))
        self.data: Union[int, float] = data
        self.leftChild: Optional[Node] = None
        self.rightChild: Optional[Node] = None

class BST(object):
    """ 
    Implementation of the Binary Search Tree (https://en.wikipedia.org/wiki/Binary_search_tree)

    All items on the left are smaller than given node &
    all on the right are larger than the given node
      
    Attributes:\n
        root: The root node of the BST Tree
        size: counter for the number of items inside the tree
    """

    def __init__(self) -> None:
        """ 
        Constructor of the BST class
        """
        self.root: Optional[Union[int, float]] = None
        self.size: int = 0

    def getSize(self) -> int:
        """ 
        Method used to return the size of the BST
        """
        return self.size

    def insert(self, data: Union[int, float]) -> None:
        """ 
        Public method used to insert data into the BST tree

        Parameters:\n
            data: the data to be added to the tree
        """
        assert (isinstance(data, int) or isinstance(data,float))

        # if root node isnt currently defined
        if not self.root:
            self.root = Node(data) # instantiate the root of BTS
            self.size += 1
        else:
            self.__insertNode(data, self.root)

    # Running time complexity of O (logN) -> Only if tree is balanced
    def __insertNode(self, data: Union[int, float], node: Node) -> None:
        """ 
        Private method used to insert data into the BST tree

        Parameters:\n
            data: the data to be added to the tree
            node: the root node of the tree
        """
        assert (isinstance(data, int) or isinstance(data,float))

        if data < node.data:
            if node.leftChild: # if the node has a left child
                # recursive call 
                self.__insertNode(data,node.leftChild)
            else:
                node.leftChild = Node(data)
                self.size += 1
        else:
            if node.rightChild: # if the node has a right child
                self.__insertNode(data,node.rightChild)
            else:
                node.rightChild = Node(data)
                self.size +=1


    def remove(self, data: Union[int, float]) -> None:
        """ 
        Public method used to remove data from the BST tree

        Parameters:\n
            data: the data to be removed from the tree
        """
        assert (isinstance(data, int) or isinstance(data,float))
        if self.root: # if the root exists -> are there nodes
            self.root = self.__removeData(data,self.root)

    # Using the predecessor method of removing data
    def __removeData(self, data: Union[int, float], node: Optional[Node]) -> Optional[Node]:
        """ 
        Private method used to remove data from the BST tree

        Parameters:\n
            data: the data to be removed to the tree
            node: the root node of the tree
        """
        assert (isinstance(data, int) or isinstance(data,float))

        if not node: # this is the base case my guy duuhh
            return node

        if data < node.data: # if smaller
            node.leftChild =  self.__removeData(data,node.leftChild)
        elif data > node.data:
            node.rightChild = self.__removeData(data,node.rightChild)
        else:
            # This means we are at the node we want to remove
            # There are 3 cases for this

            # 1: if there are no children -> Leaf Node
            if not node.leftChild and not node.rightChild:
                #print("Removing leaf node...")
                del node
                return None
            elif not node.leftChild: # no left but would have right ^
                #print("Removing node with right child...")
                rightChild = node.rightChild
                del node
                return rightChild
            elif not node.rightChild: # no right child
                #print("Removing node with left child...")
                leftChild = node.leftChild
                del node
                return leftChild
            else:
                #print("Removing node with two children...")
                predecessor = self.__getPredecessor(node.leftChild)
                node.data = predecessor.data
                # then get rid of the predecessor starting from left child
                # have to call remove node recursively because we will end up with
                # one of the two previous situations
                node.leftChild = self.__removeData(predecessor.data,node.leftChild)
        return node                 

    def __getPredecessor(self, node: Node) -> Node: # get largest value in left subtree
        """ 
        Private method to get the largest value in a subtree

        Parameters:\n
            node: the reference node used to find largest value in its subtree
        """
        if not node.rightChild:
            return node
        return self.__getPredecessor(node.rightChild)

    def getMinValue(self) -> Optional[Union[int, float]]:
        """ 
        Public method used to get the smallest value in the BST
        """
        if self.root: # if there are nodes in the BST
            return self.__getMin(self.root)
        else:
            return None

    def __getMin(self, node: Node) -> Union[int, float]:
        """ 
        Private method used to get the smallest value in the BST

        Parameters:\n
            node: the BST's root node
        """
        if node.leftChild: # if the left child node exists
            return self.__getMin(node.leftChild) # recursively call min
        else:
            return node.data

    def getMaxValue(self) -> Optional[Union[int, float]]:
        """ 
        Public method used to get the largest value in the BST
        """
        if self.root:
            return self.__getMax(self.root)
        else:
            return None

    def __getMax(self, node: Node) -> Union[int, float]:
        """ 
        Private method used to get the largest value in the BST

        Parameters:\n
            node: the BST's root node
        """
        if node.rightChild: # again if the right node exists
            return self.__getMax(node.rightChild) #recursively call func
        else:
            return node.data

    def inOrderTraversal(self) -> None:
        """ 
        Public method used to perform an inOrderTraversal
        """
        if self.root: # if the root node exists
            self.__inOrder(self.root)

    def __inOrder(self, node: None) -> None:
        """ 
        Private method used to perform an inOrderTraversal

        Parameters:\n
            node: the BST's root node
        """
        if node.leftChild: # if the left child exists
            self.__inOrder(node.leftChild) # will evetually print it, if it exists
        
        print("%d" % node.data)

        if node.rightChild:
            self.__inOrder(node.rightChild) # same thing

    def preOrderTraversal(self) -> None:
        """ 
        Public method used to perform an preOrderTraversal
        """
        if self.root: # if the root node exists
            self.__preOrder(self.root)

    def __preOrder(self, node: Node) -> None:
        """ 
        Private method used to perform an preOrderTraversal

        Parameters:\n
            node: the BST's root node
        """
        print("%d" % node.data)

        if node.leftChild: # if the left child exists
            self.__preOrder(node.leftChild) # will evetually print it, if it exists

        if node.rightChild:
            self.__preOrder(node.rightChild) # same thing

    def postOrderTraversal(self) -> None:
        """ 
        Public method used to perform an postOrderTraversal
        """
        if self.root: # if the root node exists
            self.__postOrder(self.root)

    def __postOrder(self, node: Node) -> None:
        """ 
        Private method used to perform an postOrderTraversal

        Parameters:\n
            node: the BST's root node
        """
        if node.leftChild: # if the left child exists
            self.__postOrder(node.leftChild) # will evetually print it, if it exists

        if node.rightChild:
            self.__postOrder(node.rightChild) # same thing

        print("%d" % node.data)



"""
Implementation of the Doubly Linked List

@author Chenghao Gong gongc12
"""

class Node:
    """ 
    This is a custom node class for the Doubly Linked List
      
    Attributes:\n
        data: The data to be stored in the node
        next: The next node in the linked list
        prev: The previous node in the linked list
    """
    def __init__(self, data):
        self.data = data
        self.next = None
        self.prev = None

class DoublyLinkedList:
    """ 
    The constructor for the doubly linked list
      
    Attributes:\n
        head: The head node in the linked list
    """
    def __init__(self):
        self.head = None
        self.size = 0

    """ 
    Given a reference to the head of a list and an integer, inserts a new node on the front of list

    Parameters:\n
        new_data: a new value being inserted to the list
    """
    def push(self, new_data):
        new_node = Node(new_data)
        new_node.next = self.head
        if self.head is not None:
            self.head.prev = new_node
        self.head = new_node
        self.size += 1

    """ 
    Given a reference to the head of a list and an integer, inserts a new node on the front of list

    Parameters:\n
        new_data: a new value being inserted to the list
        prev_node: the node being inserted after
    """
    def insertAfter(self, prev_node, new_data):

        if prev_node is None:
            print("the given previous node cannot be NULL")
            return
        new_node = Node(new_data)
        new_node.next = prev_node.next
        prev_node.next = new_node
        new_node.prev = prev_node
        if new_node.next:
            new_node.next.prev = new_node
        self.size += 1

    """ 
    Given a reference to the head of DLL and integer,appends a new node at the end

    Parameters:\n
        new_data: a new value being inserted to the list
    """
    def append(self, new_data):
        new_node = Node(new_data)
        if self.head is None:
            self.head = new_node
            self.size += 1
            return
        last = self.head
        while last.next:
            last = last.next
        last.next = new_node
        new_node.prev = last
        self.size += 1
        return

    """ 
    This function prints contents of linked list starting from the given node

    Parameters:\n
        node: the starting node for a given linkedlist
    """
    ## @brief a accessor for the module
    #  @details 
    #  @param self reference to the object
    #  @param 
    def printList(self, node):
        print("\nTraversal in forward direction")
        while node:
            print(" % d" % (node.data))
            last = node
            node = node.next

    ##  @brief A getter method for the linkedlist
    #   @details Return the state variable head
    #   @param self reference to the object
    #   @return the first element in the linked list

    def getFirst(self):
        return self.head

    ##  @brief A getter method for the linkedlist
    #   @details Return the state variable
    #   @param self reference to the object
    #   @return the last element in the linked list
    def getLast(self):
        cur = self.head
        while cur.next is not None:
            cur = cur.next
        return cur
    ##  @brief Check if element already present in the linked list
    #   @details if the node contains object, __eq__ method must be overwritten in the class
    #   @param self reference to the object
    #   @return boolean value indicate the result
    def contains(self,element):
        cur = self.head
        while cur is not None:
            if cur.data == element:
                return True
            cur = cur.next
        return False

    ##  @brief A getter method for the linkedlist
    #   @details Return the state variable size
    #   @param self reference to the object
    #   @return the size of the current linkedlist

    def sizeOf(self):
        return self.size

    ##  @brief A getter method for the linked list
    #   @details Return the node
    #   @param self reference to the object
    #   @param index an index specified the position
    #   @return the node in the givein index

    def get(self,index):
        if index >= self.size or index < 0:
            raise Exception("Not retriveable")
        cur = self.head
        for i in range(index):
            cur = cur.next
        return cur

    ##  @brief A setter used to change the linked list state
    #   @details remove the element in the given index
    #   @param self reference to the object
    #   @param index an index specified the position
    def remove(self, index):
        if index >= self.size or index < 0:
            raise Exception("Not retriveable")
        cur = self.head
        for i in range(index):
            cur = cur.next
        if cur != self.head:
            prev = cur.prev
            prev.next = cur.next
        else:
            self.head = cur.next













from Pypeline.Pypes.MaxHeap import MaxHeap
import numpy as np
import unittest

class MaxHeapTest(unittest.TestCase):
	"""
	Tests for methods in the MaxHeap class.
	"""

	@classmethod
	def setUpClass(cls):
		cls.li = [2, 4 , 10, 8, 90, 65, 18]
		cls.li2 = [4, 5, 10, 1]

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		self.heap = MaxHeap()
		# heap.heapify(li)
		# a = heap.heappop()
		# print(a, type(a))
		# arr = heap.to_array()
		# print(arr, type(arr))
		# heap.merge(li2)
		# print(heap)
		# heap.heappop()
		# print(heap)

	def tearDown(self):
		self.heap = None


	def test_heapify(self):
		self.heap.heapify(self.li)
		assert self.heap.MaxHeap

	def test_heappop(self):
		self.heap.heapify(self.li)
		x = self.heap.heappop()
		assert x == 2

	def test_heappush(self):
		self.heap.heapify(self.li)
		self.heap.heappush(12)
		assert self.heap.MaxHeap[-1] == 12
		[self.heap.heappop() for _ in range(3)]
		assert self.heap.heappop() == 12

	def test_to_array(self):
		self.heap.heapify(self.li)
		arr = self.heap.to_array()
		assert type(arr) == np.ndarray
		assert list(arr) == self.li

	def test_merge(self):
		self.heap.heapify(self.li)
		self.heap.merge(self.li2)
		assert len(self.li) + len(self.li2) == len(self.heap.MaxHeap)



"""
Implementation of the Min Heap

"""

import heapq
from typing import Any, List, Optional

import numpy as np

class MinHeap():
    """
    Implementation of the Min Heap (https://en.wikipedia.org/wiki/Heap_(data_structure))

    For any given node C, if P is a parent node of C, the key of P is less than or equal to the key of C

    """
    def __init__(self, li: List[int]=[]):
        self.MinHeap: List[int] = []
        assert isinstance(li, list)
        self.heapify(li)

    def __repr__(self) -> str:
        return str(self.MinHeap)

    def __str__(self) -> str:
        return str(self.MinHeap)

    def heapify(self, li: List[int]) -> None:
        assert isinstance(li, list)
        if self.MinHeap:
            li2 = [Restore(i) for i in self.MinHeap]
            self.MinHeap = [Backwards(i) for i in li2 + li]
            heapq.heapify(self.MinHeap)
        else:
            self.MinHeap = [Backwards(i) for i in li]
            heapq.heapify(self.MinHeap)

    def heappop(self) -> float:
        if not self.MinHeap:
            exception = "Index Error: Length of Heap is 0."
            raise Exception(exception)
        popped = heapq.heappop(self.MinHeap)
        return float(Restore(popped))

    def heappush(self, val: int) -> None:
        assert (isinstance(val, int))
        heapq.heappush(self.MinHeap,Backwards(val))

    def to_array(self) -> Any:
        return np.array([Restore(i) for i in self.MinHeap])

    def to_list(self) -> List[int]:
        return self.MinHeap

    def merge(self, li: List[int]) -> None:
        assert isinstance(li, list)
        self.heapify(li)

class Backwards(int):
    def __lt__(self, other: int) -> bool:
        return not int.__le__(self, other)
    def __le__(self, other: int) -> bool:
        return not int.__lt__(self, other)
    def __gt__(self, other: int) -> bool:
        return not int.__ge__(self, other)
    def __ge__(self, other: int) -> bool:
        return not int.__gt__(self, other)

class Restore(int):
    def __lt__(self, other: int) -> bool:
        return int.__le__(self, other)
    def __le__(self, other: int) -> bool:
        return int.__lt__(self, other)
    def __gt__(self, other: int) -> bool:
        return int.__ge__(self, other)
    def __ge__(self, other: int) -> bool:
        return int.__gt__(self, other)

if __name__ == "__main__":
    li = [2, 4 , 10, 8, 90, 65, 18]
    li2 = [4, 5, 10, 1]
    heap = MinHeap()
    heap.heapify(li)
    a = heap.heappop()
    print(a, type(a))
    arr = heap.to_array()
    print(arr, type(arr))
    heap.merge(li2)
    print(heap)
    heap.heappop()
    print(heap)


    


from Pypeline.Pypes.Sorting.Quicksort import quickSort, partition
import unittest


class QuicksortTest(unittest.TestCase):
    """
	Tests for functions in the Quicksort module.
	"""

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        self.testarray1 = [7, 6, 4, 3, 5, 2, 13, 8]
        self.testarray2 = [1, 4, 7, -5, 3, 5, 8, 19]

    def tearDown(self):
        pass

    def test_partition(self):
        self.assertEqual(partition(self.testarray1,0,len(self.testarray1)-1, order='ascend'),6)
        self.assertEqual(partition(self.testarray2, 0, len(self.testarray2) - 1, order='ascend'), 7)

    def test_quickSort(self):
        quickSort(self.testarray1,0,len(self.testarray1)-1)
        quickSort(self.testarray2, 0, len(self.testarray2) - 1)
        self.assertListEqual(self.testarray1, [2,3,4,5,6,7,8,13])
        self.assertListEqual(self.testarray2, [-5,1,3,4,5,7,8,19])



from Pypeline.Pypes.Stack import Stack
import numpy as np
import unittest

class StackTest(unittest.TestCase):
	"""
	Tests for methods in the Stack class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		self.s = Stack()
		pass #TODO

	def tearDown(self):
		self.s = None
		pass #TODO

	def test_isEmpty(self):
		assert self.s.isEmpty()

	def test_sizeStack(self):
		self.s.push(2)
		assert self.s.sizeStack() == 1

	def test_push(self):
		for i in range(100):
			self.s.push(i)
		assert self.s.sizeStack() == 100

	def test_pop(self):
		[self.s.push(x) for x in range(100, 0, -1)]
		assert(all([(self.s.pop() == (i+1)) for i in range(self.s.sizeStack())]))
		assert self.s.sizeStack() == 0

	def test_peek(self):
		self.s.push(20)
		assert self.s.peek() == 20

	def test_to_array(self):
		ref = []
		for x in range(10):
			self.s.push(x) 
			ref.append(x)

		assert self.s.to_list() == ref

	def test_to_list(self):
		ref = np.zeros(10, dtype=int)
		for x in range(10):
			self.s.push(x) 
			ref[x] = x

		assert (self.s.to_array() == ref).all()



from Pypeline.Pypes.BinarySearchTree import BST 
import unittest

class NodeTest(unittest.TestCase):
	"""
	Tests for methods in the Node class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO


class BSTTest(unittest.TestCase):
	"""
	Tests for methods in the BST class.
	"""

	@classmethod
	def setUpClass(cls):
		cls.vals = [10.1, 3, 20, 12, 9, 5, 30, 21]

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		self.bst = BST()

	def tearDown(self):
		self.bst = None

	def test_getSize(self):
		self.bst.insert(1)
		assert self.bst.getSize() == 1

	def test_insert(self):
		[self.bst.insert(x) for x in self.vals]
		assert self.bst.getSize() == 8

	def test_remove(self):
		[self.bst.insert(x) for x in self.vals]
		[self.bst.remove(x) for x in self.vals]
		assert not(self.bst.inOrderTraversal())

	def test_getMinValue(self):
		self.bst.insert(2)
		self.bst.insert(200)
		assert self.bst.getMinValue() == 2

	def test_getMaxValue(self):
		self.bst.insert(100)
		self.bst.insert(2)
		assert self.bst.getMaxValue() == 100

	def test_inOrderTraversal(self):
		# print(self.vals)
		# [self.bst.insert(x) for x in self.vals]
		# lst = self.bst.inOrderTraversal() 
		# assert lst == sorted(self.vals)
		pass # Can't really test this or the following as they interact with the console output.

	def test_preOrderTraversal(self):
		print(self.bst.preOrderTraversal())
		pass
		# raise NotImplementedError() #TODO: test preOrderTraversal

	def test_postOrderTraversal(self):
		print(self.bst.postOrderTraversal())
		pass
		# raise NotImplementedError() #TODO: test postOrderTraversal



from Pypeline.Pypes.Sorting.Shellsort import shellSort
import unittest

class ShellsortTest(unittest.TestCase):
	"""
	Tests for functions in the Shellsort module.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO

	def test_shellSort(self):
		arr1 = [1,4,6,-5,1]
		arr2 = [7,3,5,2,6,8,1,5,6,10]
		shellSort(arr1)
		shellSort(arr2)
		self.assertListEqual(arr1, [-5, 1, 1, 4, 6])
		self.assertListEqual(arr2, [1, 2, 3, 5, 5, 6, 6, 7, 8, 10])



from Pypeline.Pypes.StringSearch.KMP import *
import unittest

class KMPTest(unittest.TestCase):
	"""
	Tests for functions in the KMP module.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO

	def test_KMPSearch(self):
		pattern1 = "txt1"
		txt1 = "tfsdafsadasdtxt1sadsa"
		pattern2 = " "
		pattern3 = "sad"
		txt2 = " "
		self.assertEqual(KMPSearch(pattern1,txt1),12)
		self.assertEqual(KMPSearch(pattern2,txt1),None)
		self.assertEqual(KMPSearch(pattern3,txt1),6)
		self.assertEqual(KMPSearch(pattern1,txt2),None)
		





"""
Implementation of the Ternay Search Tree

"""

from typing import Any, Optional

class Node(object):
    """ 
    This is a custom node class for the Doubly Linked List
      
    Attributes:\n
        char: The char to be stored
        leftNode: The left node child
        rightNode: The right node child
        middleNode: The middle node child
        value: The value associated with the char
    """
    def __init__(self, char: str):
        self.char: str = char
        self.leftNode: Optional[Any] = None
        self.rightNode: Optional[Any] = None
        self.middleNode: Optional[Node] = None
        self.value: int = 0

class TST(object):
    """
    Implementation of the Ternay Search Tree (https://en.wikipedia.org/wiki/Ternary_search_tree)

    A type of trie (sometimes called a prefix tree) where nodes are arranged in a manner similar to a binary search tree,
    but with up to three children rather than the binary tree's limit of two.

    """
    def __init__(self) -> None:
        self.rootNode = None

    def put(self, key: str, value: int):
        assert (isinstance(key, str))
        self.rootNode = self.putItem(self.rootNode, key, value, 0)

    def putItem(self, node: Optional[Node], key: int, value: Any, index: int) -> Node: # key -> the string; index -> index of string
        c = key[index]

        if node == None:
            node = Node(c) # update the given node

        if c < node.char:
            # dont increment index yet, as in the example, it still has to be considered
            node.leftNode = self.putItem(node.leftNode, key,value,index)

        elif c > node.char:
            # dont increment index yet, as in the example, it still has to be considered
            node.rightNode = self.putItem(node.rightNode, key,value,index)

        elif index < len(key) -1: # and not and the end of the string
            # increment cause we will be going to the middle child through the node list
            index += 1
            node.middleNode = self.putItem(node.middleNode, key, value, index)
        else: # now we are at the last index of the string so we can update the value
            node.value = value

        return node

    def get(self, key: int) -> int:

        node = self.getItem(self.rootNode, key, 0) 

        if node == None:
            return -1 # the key is not in the tree

        return node.value

    def getItem(self,node: Optional[Node], key: str, index: int) -> Optional[Node]:
        if node == None:
            return node

        c = key[index]

        # recursive implementation
        if c < node.char: # go left
            return self.getItem(node.leftNode, key, index)
        elif c > node.char: # go right
            return self.getItem(node.rightNode, key, index)
        elif index < len(key) -1: # if we are not at the strings lengt
            index += 1
            return self.getItem(node.middleNode, key, index)
        else: # we would be at the last node at this point
            return node
        
if __name__ == "__main__":
    tst = TST()

    tst.put("cat",5)
    tst.put("dog",30)
    tst.put("jack",19)
    tst.put("robby",12)

    print("The value of cat is: {}".format(tst.get("cat")))
    print("The value of robby is: {}".format(tst.get("robby")))

    print("Updating...")
    tst.put("robby",27)
    print("The value of cat is: {}".format(tst.get("roby")))



"""
Implementation of the Queue Data Stucture

# FIFO Structure -> First in First Out

"""

from typing import Any, Optional, List

import numpy as np

class Queue:
    def __init__(self, maxlen: int=1e9):
        self.queue: List[Any] = []
        self.maxlen: int = maxlen

    def isEmpty(self) -> bool:
        return self.queue == []

    def sizeQueue(self) -> int:
        return len(self.queue)

    # # Main Methods: Enqueue Dequeue Peel
    def enqueue(self, data: Any) -> None:
        if self.sizeQueue() < self.maxlen: 
            self.queue.append(data)
        else:
            _ = self.dequeue()
            del _
            self.queue.append(data)

    def dequeue(self) -> Any:
        data = self.queue[0]
        del self.queue[0]
        return data

    def peek(self) -> Any:
        return self.queue[0]

    def to_array(self) -> Any:
        return np.array(self.queue)

    def to_list(self) -> List[int]:
        return self.queue



from setuptools import setup

setup(
    name = "python-pypeline",
    url = "https://github.com/akiljames83/Pypeline",
    version = "1.0.0",
    install_requires = ['docutils>=0.3','numpy>=0.1'],

    # metadata for upload to PyPI
    author = "Akil Hamilton",
    packages=['Pypeline', 'Pypeline.Pypes'],
    author_email = "akil.james83@gmail.com",
    description = "This is a package which will allow for easy implementation of basic Data Structures.",
    license = "MIT",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords = "data structures implementation"
)


from Pypeline.Pypes import RedBlackBST
import unittest

class RedBlackBSTTest(unittest.TestCase):
	"""
	Tests for functions in the RedBlackBST module.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO

	def test_isRed(self):
		node = RedBlackBST.Node(key=1, val=3, size=0, color=True)
		assert RedBlackBST.isRed(node) == True

	def test_size(self):
		node = RedBlackBST.Node(key=1, val=3, size=0, color=True)
		assert RedBlackBST.size(node) == 0

	def test_rotateLeft(self):
		pass

	def test_rotateRight(self):
		pass

	def test_flipColors(self):
		pass

	def test_cmp(self):
		pass

class NodeTest(unittest.TestCase):
	"""
	Tests for methods in the Node class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO


class redBlackBSTTest(unittest.TestCase):
	"""
	Tests for methods in the redBlackBST class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO

	def test_put(self):
		tree = RedBlackBST.redBlackBST()
		tree.put(1, 3)
		assert tree.root.val == 3

	def test_actualPut(self):
		tree = RedBlackBST.redBlackBST()
		tree.root = tree.actualPut(tree.root, 1, 3)
		assert tree.root.val == 3



from Pypeline.Pypes.Queue import Queue
import numpy as np
import unittest

class QueueTest(unittest.TestCase):
	"""
	Tests for methods in the Queue class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		self.q = Queue()

	def tearDown(self):
		self.q = None

	def test_isEmpty(self):
		assert self.q.isEmpty()

	def test_sizeQueue(self):
		assert self.q.sizeQueue() == 0

	def test_enqueue_deque(self):
		[self.q.enqueue(i) for i in range(100)]
		self.q.enqueue('TEST')
		self.q.enqueue(list(range(5)))
		assert all([self.q.dequeue() == i for i in range(100)])
		assert self.q.dequeue() == 'TEST'
		
		assert self.q.dequeue() == list(range(5))

	def test_peek(self):
		self.q.enqueue('TEST')
		assert self.q.peek() == 'TEST'

	def test_to_array(self):
		ref = np.zeros(10, dtype=int)
		for x in range(10):
			self.q.enqueue(x) 
			ref[x] = x
		assert (self.q.to_array() == ref).all()

	def test_to_list(self):
		ref = []
		for x in range(10):
			self.q.enqueue(x) 
			ref.append(x)

		assert self.q.to_list() == ref



from Pypeline.Pypes import TernarySearchTree
import unittest

class NodeTest(unittest.TestCase):
	"""
	Tests for methods in the Node class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO


class TSTTest(unittest.TestCase):
	"""
	Tests for methods in the TST class.
	"""

	@classmethod
	def setUpClass(cls):
		pass #TODO

	@classmethod
	def tearDownClass(cls):
		pass #TODO

	def setUp(self):
		pass #TODO

	def tearDown(self):
		pass #TODO

	def test_put(self):
		tree = TernarySearchTree.TST()
		tree.put('a', 5)
		assert tree.rootNode.char == 'a'
		assert tree.rootNode.value == 5

	def test_putItem(self):
		tree = TernarySearchTree.TST()
		n = tree.putItem(tree.rootNode, 'a', 5, 0)
		assert n.char == 'a'
		assert n.value == 5

	def test_get(self):
		tree = TernarySearchTree.TST()
		tree.put('a', 5)
		assert tree.get('a') == 5

	def test_getItem(self):
		tree = TernarySearchTree.TST()
		tree.put('a', 5)
		n = tree.getItem(tree.rootNode, 'a', 0)
		assert n.char == 'a'
		assert n.value == 5



"""
Implementation of the Max Heap

"""

import heapq
from typing import Any, List, Optional, Union

import numpy as np

class MaxHeap():
    """
    Implementation of the Max Heap (https://en.wikipedia.org/wiki/Min-max_heap)

    For any given node C, if P is a parent node of C, then the key (the value) of P is greater than or equal to the key of C

    """
    def __init__(self, li: List[Union[float, int]]=[]) -> None:
        self.MaxHeap = []
        assert isinstance(li, list)
        self.heapify(li)
            

    def __repr__(self) -> str:
        return str(self.MaxHeap)

    def __str__(self) -> str:
        return str(self.MaxHeap)

    def heapify(self, li: List[Union[float, int]]) -> None:
        assert isinstance(li, list)
        if self.MaxHeap:
            self.merge(li)
        else:
            self.MaxHeap = li
            heapq.heapify(self.MaxHeap)

    def heappop(self) -> Any:
        if not self.MaxHeap:
            exception = "Index Error: Length of Heap is 0."
            raise Exception(exception)
        return heapq.heappop(self.MaxHeap)

    def heappush(self, val: Union[float, int]) -> None:
        assert (isinstance(val, float) or isinstance(val, int))
        heapq.heappush(self.MaxHeap,val)

    def to_array(self) -> Any:
        return np.array(self.MaxHeap)

    def merge(self, li: List[Union[float, int]]) -> None:
        assert isinstance(li, list)
        self.MaxHeap = self.MaxHeap + li
        heapq.heapify(self.MaxHeap)

    